"""
Copyright 2024 EMD International
License for this script: MIT https://opensource.org/license/mit/
License for windPRO commercial software: https://www.emd-international.com/contact-us/general-terms-conditions-sale/
"""

import os
from datetime import datetime
from windproapi.utils import get_windpro_sample_path
from windproapi import WindProApi
from windproapi import nan_to_skipvalue

# Opening windPRO
_windproapi = WindProApi()
working_dir = os.path.join(get_windpro_sample_path('4.1'), 'New Salem', '4.1')
project_path = os.path.join(working_dir, 'New Salem.w41p')

_windproapi.start_windpro_random_port()

# Services
obj_wtg_service = _windproapi.get_service('ObjWtgService')
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
wtg_explorer_service = _windproapi.get_service('WtgExplorerService')
calculation_service = _windproapi.get_service('CalculationService')
calc_park_service = _windproapi.get_service('CalcParkService')
factory = _windproapi.get_factory('WindproService')

# Loading New Salem project
project_service.LoadFromFile(filename=project_path)

# Getting wind turbine generators already present
objs = objects_service.GetObjects(apiObjType='NewWTG')
wtg_obj = obj_wtg_service.GetWtgObject(objs[0].Handle)
print(wtg_obj)


## Curtailment helper functions
# Function for changing the date time into format for windPRO (same as excel)
def excel_date(date1):
    temp = datetime(1899, 12, 30)    # Note, not 31st Dec but 30th!
    delta = date1 - temp
    return float(delta.days) + (float(delta.seconds) / 86400)

# For chaning curtailments by type
def change_curtailment(curtailment, curtailment_type, condition_from, condition_to):
    for c in curtailment.CurtailmentConditions.TApiWtgCurtailmentCondition:
        if c.CurtailmentCondition == curtailment_type:
            c.ConditionFrom = condition_from
            c.ConditionTo = condition_to
            return
    raise KeyError(f'Could not find {curtailment_type}.')

def delete_curtailment(curtailment, curtailment_type):
    for i, c in enumerate(curtailment.CurtailmentConditions.TApiWtgCurtailmentCondition):
        if c.CurtailmentCondition == curtailment_type:
            curtailment.CurtailmentConditions.TApiWtgCurtailmentCondition.pop(i)
            return
    raise KeyError(f'Could not find {curtailment_type}.')

## Adding predefined curtailment
# Curtailment template for birds
newCurtailment = obj_wtg_service.GetCurtailmentTemplate('ApiCTBirds')
# Adjusting curtailment
newCurtailment.Name = 'Bird from default'
delete_curtailment(newCurtailment, 'ApiCCTemperature')
delete_curtailment(newCurtailment, 'ApiCCWindDirection')
delete_curtailment(newCurtailment, 'ApiCCWindSpeed')
# Time of day is set by fraction of 24 h values
change_curtailment(newCurtailment, 'ApiCCTime', 5/24, 7/24)
# Dates need to be set as days since 30th December 1899 like in excel. The year does not matter but needs to be given.
change_curtailment(newCurtailment, 'ApiCCDate', excel_date(datetime(2020, 3, 1)), excel_date(datetime(2020, 5, 1)))
# Times before sunset and sunrise are set by integer values.
# A list is available from the wind scripting api
print(obj_wtg_service.GetSunriseAndSunsetValues())
# using '0:45h before sunset: 19955', '0:45h after sunset: 20045'
change_curtailment(newCurtailment, 'ApiCCSunriseSunset', 19955, 20045)

wtg_obj = obj_wtg_service.GetWtgObject(objs[0].Handle)
# Adding curtailment list if None is present
if wtg_obj.Curtailment is None:
    TApiWtgCurtailments = factory.TApiWtgCurtailments()
    wtg_obj.Curtailment = TApiWtgCurtailments

wtg_obj.Curtailment.TApiWtgCurtailment.append(newCurtailment)
nan_to_skipvalue(wtg_obj)
obj_wtg_service.SetWtgObject(wtg_obj)

## Adding Curtailment with reduction in the mode
# Finding a power curve
print(wtg_explorer_service.GetWtgFromFile(filename=wtg_obj.Filename, details=True))

# Using similar settings as above
change_curtailment(newCurtailment, 'ApiCCDate', excel_date(datetime(2020, 8, 1)), excel_date(datetime(2020, 10, 1)))
newCurtailment.OperationMode = 'ApiCOMPowerCurve'
newCurtailment.ActionUID = '{6B20F289-1E7F-430D-AD71-C3A2D617ED38}'
newCurtailment.Name = 'Scripting reduced power curve'

wtg_obj = obj_wtg_service.GetWtgObject(objs[0].Handle)
# Adding curtailment list if None is present
if wtg_obj.Curtailment is None:
    TApiWtgCurtailments = factory.TApiWtgCurtailments()
    wtg_obj.Curtailment = TApiWtgCurtailments

wtg_obj.Curtailment.TApiWtgCurtailment.append(newCurtailment)
nan_to_skipvalue(wtg_obj)
obj_wtg_service.SetWtgObject(wtg_obj)

## Adding freely defined curtailment
newCurtailment = obj_wtg_service.GetCurtailmentTemplate('ApiCTOther')
newCurtailment.Name = f'Freely defined from Scripting'

# Making conditions
condition = factory.TApiWtgCurtailmentCondition()
condition.CurtailmentCondition = 'ApiCCDateTime'
condition.ConditionFrom = excel_date(datetime(2024, 8, 1))
condition.ConditionTo = excel_date(datetime(2020, 8, 31))

TApiWtgCurtailmentConditions = factory.TApiWtgCurtailmentConditions()
TApiWtgCurtailmentConditions.TApiWtgCurtailmentCondition.append(condition)

newCurtailment.CurtailmentConditions = TApiWtgCurtailmentConditions

wtg_obj = obj_wtg_service.GetWtgObject(objs[0].Handle)
# Adding curtailment list if None is present
if wtg_obj.Curtailment is None:
    TApiWtgCurtailments = factory.TApiWtgCurtailments()
    wtg_obj.Curtailment = TApiWtgCurtailments

wtg_obj.Curtailment.TApiWtgCurtailment.append(newCurtailment)
nan_to_skipvalue(wtg_obj)
obj_wtg_service.SetWtgObject(wtg_obj)