"""
Copyright 2024 EMD International
License for this script: MIT https://opensource.org/license/mit/
License for windPRO commercial software: https://www.emd-international.com/contact-us/general-terms-conditions-sale/
"""

import os
import time
from windproapi import WindProApi, nan_to_skipvalue
from windproapi.utils import get_windpro_sample_path
import numpy as np
import pandas as pd

# Opening windPRO
_windproapi = WindProApi()
working_dir = os.path.join(get_windpro_sample_path('4.1'), 'Aparados da Serra', '4.1')
project_path = os.path.join(working_dir, 'Aparados da Serra basic project.w41p')

_windproapi.start_windpro_random_port()

# Services
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
calculation_service = _windproapi.get_service('CalculationService')
optimize_service = _windproapi.get_service('OptimizeService')
wtg_explorer_service = _windproapi.get_service('WtgExplorerService')
factory = _windproapi.get_factory('WindproService')

# Loading project file
project_service.LoadFromFile(filename=project_path)

# Setting up an optimization for multiple sizes as potential layouts.
# Choosing range of number of turbines to be investigated and a resource grid as the basis
N_turbines = np.arange(1, 5, 1)
resource_file = os.path.join(working_dir, 'Aparados da Serra basic project_Hub_98.0_0.rsf')

# Making new session
session_name = 'New Session from scripting'
session_guid = optimize_service.AddSession(session_name)
optimize_service.OpenSession(session_guid)

# Adding a new site to the calculations
site_uid = optimize_service.AddSite('Test Fill Max')
site = optimize_service.GetSite(site_uid)

# Setting the site specific values.
# We stay with the default onshore wake decay constant which is default behaviour
wtg_area_obj = objects_service.GetObjects('WTGareas')[0]
site.WTGAreaHandle = wtg_area_obj.Handle
site.WindResourceFiles.string[0] = resource_file
nan_to_skipvalue(site)
optimize_service.SetSite(site)

# Adding specific wind turbine
wtg_from_explorer = wtg_explorer_service.GetWtgFromUID(uid='{832CAEAF-54E2-4A29-842C-144AA88B553E}')
# Setting default tower height to match the rsf file
# wtg_from_explorer.DefHubHeight = 98.0
wtg_uid = optimize_service.AddWTG(siteGuid=site_uid,
                                  name=os.path.basename(wtg_from_explorer.FileName),
                                  wtgFile=wtg_from_explorer.FileName)
wtg = optimize_service.GetWTG(wtg_uid)
wtg.Hubheight = 98.0
nan_to_skipvalue(wtg)
optimize_service.SetWTG(wtg)
wtg = optimize_service.GetWTG(wtg_uid)

# Adding different turbine sizes
setup = optimize_service.GetLayoutDefaultSetup()
setup.StartModel = 'optFillStartModel'
setup.StepModel = 'optRandomLocalStepModel'
setup.MaxMinutes = 1
setup.NumberOfTurbines = 12
nan_to_skipvalue(setup)
optimize_service.CreateLayout(wtgGuid=wtg_uid, Setup=setup)

wtg = optimize_service.GetWTG(wtg_uid)

layout_list = factory.TApiStringArray()
# Getting the Size to get the Guid for the layout for starting the optimization later
new_size = optimize_service.GetSize(wtg.RunList.TApiOptimizeNameGuid[0].Guid)
layout_list.string.append(new_size.LayoutList.TApiOptimizeNameGuid[0].Guid)

# Starting runs
startRuns = optimize_service.StartRuns(layout_list)
# Runs are running in the background in the optimizer.
# Therefore, we need to check the status of all
is_running = True
while is_running:
    is_running = False
    calcs_running = 0
    time.sleep(10)
    for layout_guid in layout_list.string:
        status = optimize_service.GetStatus(layout_guid)
        if status == 'olsApiRunning':
            is_running = True
            calcs_running += 1
    print(f'Number of calculations running: {calcs_running}')


optimize_service.CloseSession(True)
