"""
Calculating the marginal energy production for a PARK calculation.
Results are stored in a sub folder called 'Scripting' in the directory of the windPRO file.

args:
1) name of the PARK calculation (does not need to be complete name, only unique string)
"""

"""
Copyright 2024 EMD International
License for this script: MIT https://opensource.org/license/mit/
License for windPRO commercial software: https://www.emd-international.com/contact-us/general-terms-conditions-sale/
"""

import os
import sys
import time

import pandas as pd
from copy import deepcopy
from windproapi.utils import get_windpro_sample_path
from windproapi import nan_to_skipvalue
from windproapi import WindProApi

import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')

def read_park_results(path_result):
    # Reading the files and
    # try:
    #     # CAUTION: this import can fail depending on the setup of the computer to use , or . as decimal seperators.
    #     df = pd.read_csv(path_result, sep=';',thousands='.',decimal=',', header=1, skiprows=[2],
    #                      usecols=['Row data/Description', 'Free mean wind speed', 'Result'],
    #                      dtype={'Row data/Description': str, 'Free mean wind speed': float, 'Result': float})
    #
    # except:
    #     pass

    try:
        df = pd.read_csv(path_result, sep=';',thousands=',',decimal='.', header=1, skiprows=[2],
                         usecols=['Row data/Description', 'Free mean wind speed', 'Result'],
                         dtype={'Row data/Description': str, 'Free mean wind speed': float, 'Result': float},
                         encoding='cp1252')
    except:
        raise NotImplementedError('Other encodings not implemented')
    df.index = df['Row data/Description']
    return df
def get_obj_from_list(objs, variable, value, exact_match=True):
    """Exact match"""
    l_obj = []
    for o in objs:
        if exact_match:
            if o[variable] == value:
                l_obj.append(o)
        else:
            if value in o[variable]:
                l_obj.append(o)

    if len(l_obj) == 0:
        raise ValueError(f'Could not find {value} in any {variable}.')
    elif len(l_obj) == 1:
        return l_obj[0]
    else:
        raise ValueError(f'More than one variable {variable} has value {value}.')

# Opening windPRO
_windproapi = WindProApi(api_port=int(sys.argv[1]))

# Services
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
calculation_service = _windproapi.get_service('CalculationService')
calc_park_service = _windproapi.get_service('CalcParkService')
obj_wtg_service = _windproapi.get_service('ObjWtgService')
factory = _windproapi.get_factory('WindproService')

# directory for scripting stuff
working_dir = os.path.join(os.path.dirname(project_service.GetProjectFilename()), 'Scripting', 'marginal_AEP')
os.makedirs(working_dir, exist_ok=True)

# Getting all PARK calculations in
all_park_calcs = calculation_service.GetCalcs(calcType='CalcPark')

# Choosing one of the PARK calculation, by name
calc_sel = get_obj_from_list(all_park_calcs, 'Name', sys.argv[2], exact_match=False)
Handle = calc_sel.Handle
try:
    # Make a copy to modify it
    cloneHandle = calculation_service.Clone(handle=Handle)
    calculation_service.OpenCalcForEdit(handle=cloneHandle)
    clonedParkCalc = calc_park_service.GetParkCalc()
    clonedParkCalc.Name = 'SCRIPTING: for marginal energy calculation'
    nan_to_skipvalue(clonedParkCalc)
    calc_park_service.SetParkCalc(clonedParkCalc)
except Exception as e:
    raise e
finally:
    calculation_service.CloseCalc(save=True)

calculation_service.Calculate(handle=cloneHandle)

# Exporting results from the wind farm
path_result = os.path.join(working_dir, 'park_result.csv')
res_to_file = calculation_service.GetResToFiles(handle=Handle)
print(f'Result {res_to_file}')
calculation_service.ResultToFile(handle=Handle, id='Park result', filename=path_result)

df_ref = read_park_results(path_result)
print(df_ref[['Free mean wind speed', 'Result', 'Row data/Description']])

# Making park calculations ommiting single turbines
turbine_list = clonedParkCalc.NewWtgs.TApiWtgId

dict_df = {}
# Looping over all turbines and
for i in range(len(turbine_list)):
    from time import sleep
    sleep(2)
    # List of turbines to include
    temp_turbine_list = deepcopy(turbine_list)
    temp_turbine_list.pop(i)

    # Getting user description to identify turbine that is ommitted later.
    obj = objects_service.GetObjectFromHandle(handle=turbine_list[i].Handle)
    label_ommited = obj.UserDescription

    # Opening for edit
    try:
        calculation_service.OpenCalcForEdit(handle=cloneHandle)
        temp_park_calc = calc_park_service.GetParkCalc()

        # Changing number of turbines in the calculation
        temp_park_calc.NewWtgs.TApiWtgId = temp_turbine_list
        nan_to_skipvalue(temp_park_calc)

        # Setting the with new settings to park calculation and calculating results
        calc_park_service.SetParkCalc(temp_park_calc)
    except Exception as e:
        raise e
    finally:
        calculation_service.CloseCalc(save=True)
    calculation_service.Calculate(handle=cloneHandle)

    # Exporting results from the wind farm
    path_result = os.path.join(working_dir, f'park_result_{i}.csv')
    calculation_service.ResultToFile(handle=cloneHandle, id='Park result', filename=path_result)

    # Reading the files and
    try:
        import pandas as pd

        # CAUTION: this can fail depending on the setup of the computer to use , or . as decimal seperators.
        # Or depending on your language setup the names of the columns might be different.
        df = read_park_results(path_result)

        print(f'Calculating for turbines:')
        print(df['Row data/Description'].to_list())
        print('Total enegry: {}'.format(df['Result'].sum()))
        dict_df[label_ommited] = df['Result']
    except:
        print('Could not read results file.')

# Concatenating all productions. Production turbine as index, omitted turbine as rows.
df_all = pd.DataFrame(dict_df)

## Comparing with case where all turbines are present
# Reference power production of all wind turbines
refenrence_power = df_ref['Result'].sum()

# What is the marginal power added by each turbine
s_diff_marginal = (refenrence_power - df_all.sum(axis=0)).rename('AEP_marginal')
df_mar = pd.concat([df_ref['Result'].rename('AEP'), s_diff_marginal], axis=1)
df_mar['AEP_diff'] = df_mar['AEP'] - df_mar['AEP_marginal']
df_mar.to_csv(os.path.join(working_dir, f'marginal_AEP_{sys.argv[2]}.csv'))

print('Marginal power:')
print(df_mar)

fig, axes = plt.subplots(2,1, sharex=True)
df_mar['AEP'].plot(ax=axes[0], label='without wake')
df_mar['AEP_marginal'].plot(ax=axes[0], label='with wake')
df_mar['AEP_diff'].plot(ax=axes[1], label='Difference', c='r')
fig.suptitle('Marginal power by turbine')
axes[1].set_xlabel('Turbine [-]')
axes[0].set_ylabel('Power [MWh]')
axes[1].set_ylabel('Power [MWh]')
fig.legend()
fig.savefig(os.path.join(working_dir, f'marginal_AEP_{sys.argv[2]}.png'))
